"""In a typical workflow, these objects are automatically created by initializing the ["ASetup"][asmu.AFile] class.
The objects can be accessed by the attributes ["interface"][asmu.ASetup.interface], ["inputs"][asmu.ASetup.inputs] and ["outputs"][asmu.ASetup.outputs].
This page is an overview of the attributes one can set/get for the different objects."""
import numpy as np
import sounddevice as sd
from typing import List, Tuple, Callable
import threading

def query_devices() -> sd.DeviceList:
        """List all connected audio devices.

        Returns:
            sd.DeviceList: A DeviceList containing one dictionary for each available device.
        """
        return sd.query_devices()

class IO():
    def __init__(self, io_setup: dict) -> None:
        """Parent class for Input and Output configuration, storing various attributes.

        Args:
            io_setup (dict): ASetup input or output settings.
        """
        self._io_setup = io_setup

    def __repr__(self):
        return self.name

    @property
    def name(self) -> str:
        """IO name."""
        return self._io_setup["name"]

    @property
    def channel(self) -> int:
        """IO channel number (starting from 1)."""
        return int(self._io_setup["channel"])

    @property
    def gain(self) -> float:
        """The gain setting of the audio interface in dB."""
        return int(self._io_setup["gain"])
    @gain.setter
    def gain(self, value: float) -> None:
        self._io_setup["gain"] = float(value)
    
    @property
    def latency(self) -> int:
        """IO channel specific latency (relative to system latency) in samples."""
        try:
            return int(self._io_setup["latency"])
        except KeyError:
            return 0
    @latency.setter
    def latency(self, value: int) -> None:
            self._io_setup["latency"] = int(value)

    @property
    def color(self) -> str:
        """IO color in HEX format."""
        return self._io_setup["color"]
    @color.setter
    def color(self, value: str) -> None:
        self._io_setup["color"] = str(value)
    
    @property
    def pos(self) -> np.ndarray:
        """IO spacial coordinates in [1, 2, 3]."""
        return np.array(self._io_setup["pos"])
    @pos.setter
    def pos(self, value: np.ndarray) -> None:
        self._io_setup["pos"] = value.tolist()

    @property
    def reference(self) -> bool:
        """Reference flag, if no flag is found return `False`."""
        try:
            return bool(self._io_setup["reference"])
        except KeyError:
            return False
    @reference.setter
    def reference(self, value: bool) -> None:
        if value:
            self._io_setup["reference"] = True
        else:
            try:
                del self._io_setup["reference"]
            except KeyError:
                pass

    @property
    def io_setup(self):
        """IO setup dictionary used in ASetup files."""
        return self._io_setup
    
    # CALIBRATION FACTORS - defined by amplitude -> direct signal conversion!
    @property
    def cPa(self) -> float:
        """Calibration factor to covert the arbitrary unit to Pascal (Pa)."""
        return float(self._io_setup["cPa"])
    @cPa.setter
    def cPa(self, value: float) -> None:
        self._io_setup["cPa"] = float(value)

    @property
    def fPa(self) -> float:
        """Calibration frequency of cPa calibration."""
        return int(self._io_setup["fPa"])
    @fPa.setter
    def fPa(self, value: float) -> None:
        self._io_setup["fPa"] = float(value)

    @property
    def cV(self) -> float:
        """Calibration factor to covert the arbitrary unit to Volt (V)."""
        return float(self._io_setup["cV"])
    @cV.setter
    def cV(self, value: float) -> None:
        self._io_setup["cV"] = float(value)

    @property
    def fV(self) -> float:
        """Calibration frequency of cV calibration."""
        return float(self._io_setup["fV"])
    @fV.setter
    def fV(self, value: float) -> None:
        self._io_setup["fV"] = float(value)

    @property
    def cFR(self) -> np.ndarray:
        """Discrete frequency response calibration vector."""
        return self._io_setup["cFR"]
    @cFR.setter
    def cFR(self, value: np.ndarray) -> None:
        self._io_setup["cFR"] = np.array(value)

    @property
    def fFR(self) -> np.ndarray:
        """Discrete frequency response frequency vector."""
        return self._io_setup["fFR"]
    @fFR.setter
    def fFR(self, value: np.ndarray) -> None:
        self._io_setup["fFR"] = np.array(value)


class Input(IO):
    def __init__(self, input_setup: dict) -> None: 
        """Input class, typically automatically created when ASetup is initialized.

        Args:
            input_setup (dict): ASetup input settings.
        """
        self._input_setup = input_setup
        super().__init__(self._input_setup)


class Output(IO):
    def __init__(self, output_setup: dict) -> None: 
        """Output class, typically automatically created when ASetup is initialized.

        Args:
            output_setup (dict): ASetup output settings.
        """
        self._output_setup = output_setup
        super().__init__(self._output_setup)


class Interface():
    def __init__(self, interface_setup: dict) -> None:
        """Class for audio interface communication and settings.

        Args:
            interface_setup (dict): ASetup interface settings.
        """
        self._interface_setup = interface_setup

    @property
    def name(self) -> None:
        """Interface name"""
        return self._interface_setup["name"]

    @property
    def device(self):
        """Device name obtained by [`queury_devices()`][asmu.io.query_devices]."""
        return self._interface_setup["device"]
    @device.setter
    def device(self, value):
        self._interface_setup["device"] = value

    @property
    def samplerate(self):
        """Samplerate"""
        return int(self._interface_setup["samplerate"])

    @property
    def blocksize(self) -> int:
        """Blocksize to use for interface communication (should be a power of 2).
            - Lower blocksize -> Reduces latency, but increases CPU load.
            - Higher blocksize -> Increases latency, but reduces CPU load.
        Typical values are between 32 and 8192."""
        return int(self._interface_setup['blocksize'])
    
    @property
    def latency(self) -> int:
        """System latency in samples."""
        try:
            return int(self._interface_setup["latency"])
        except KeyError:
            return None
    @latency.setter
    def latency(self, value: int) -> None:
        self._interface_setup["latency"] = int(value)

    def cal_latency(self, time):
        """Calculate and store loopback latency without physical connection.

        !!! warning
            Dont rely on this method, as it only calculates the ADC/DAC's internal latency. 
            Use [latency_from_rec.py](../examples.md/#latency_from_recpy) to compare this result with the real loopback calibration.

        Args:
            time (CData): The time object given in the Interface callback function.
        """
        self.latency = round((time.outputBufferDacTime-time.inputBufferAdcTime)*self.samplerate + 1) # the +1 was measured experimentally (could be the cable?)

    # SOUNDDEVICE 
    def _init_sounddevice(self, io:Tuple[List[Input], List[Output]], stream) -> None:
        """Initiializes sounddevice with the classes attributes for the given lists of inputs and outputs.
        !!! warning
            This method is currently only implemented for ASIO devices (Windows).
    
        Args:
            io (tuple of lists): Tuple that contains lists of Input and Output objects.
                If one of them is not needed, leave list empty.

        Raises:
            AttributeError: No audio device specified
        """
        stream.samplerate = self.samplerate
        stream.dtype = np.float32
        stream.blocksize = self.blocksize
        try:
            stream.device = self.device # TODO - different input and output device
        except KeyError:
            raise AttributeError("No audio device specified")

        if "ASIO" in self.device:
            if io[0]:
                in_channels = [input.channel - 1 for input in io[0]] # convert to channel names starting with 0
                asio_in = sd.AsioSettings(channel_selectors=in_channels)

                if not io[1]:
                    stream.extra_settings = asio_in
                    stream.channels = len(in_channels)
                    return

            if io[1]:
                out_channels = [output.channel - 1 for output in io[1]]
                asio_out = sd.AsioSettings(channel_selectors=out_channels)

                if not io[0]:
                    stream.extra_settings = asio_out
                    stream.channels = len(out_channels)
                    return
            
            if io[0] and io[1]:
                stream.extra_settings = (asio_in, asio_out)
                stream.channels = (len(in_channels), len(out_channels))
                return
            
        elif "CoreAudio" in self.device: 
            raise NotImplementedError
            inChannels = [c - 1 for c in channels[0]] # convert to channel names starting with 0
            ca_in = sd.CoreAudioSettings(channel_map=inChannels)

            if channels[1]:
                outChannels = [-1]*sd.query_devices(device=self.device, kind="output")["max_output_channels"]
                for idx, c in enumerate(channels[1]):
                    outChannels[c-1] = idx
                ca_out = sd.CoreAudioSettings(channel_map=outChannels)

                stream.extra_settings = (ca_in, ca_out)
                stream.channels = (len(inChannels), len(outChannels))
            else:
                stream.extra_settings = ca_in
                stream.channels = len(inChannels)
        else:
            raise NotImplementedError
        
    def callback(self, indata: np.ndarray, outdata: np.ndarray, frames: int, time):
        """Sounddevice callback function that has to be implemented by the user.
            This function is used for realtime processing auf audio signals in the form of numpy arrays.

        Args:
            indata (np.ndarray): Data optained by the interface (blocksize x channels)
            outdata (np.ndarray): Data written to the interface (blocksize x channels)
            frames (int): blocksize
            time (_type_): CDate time object to process ADC/DAC timings (used to compute latency).

        Raises:
            NotImplementedError: No callback implemnented by user
        """
        raise NotImplementedError("No callback implemnented by user")
                
    def _callback(self, indata: np.ndarray, outdata: np.ndarray, frames: int, time, status) -> None:
        if status:
            print(status, flush=True)

        outdata.fill(0)
        self.callback(indata, outdata, frames, time)

    def stream(self, io:Tuple[List[Input], List[Output]], finished_callback: Callable=None) -> sd.Stream:
        """Initiializes sounddevice with the classes attributes for the given lists of inputs and outputs, then returns the sounddevice.Stream to be used in a with statement tos tart the stream.

        !!! warning
            Streams are currently only working for ASIO devices (Windows).

        Args:
            io (tuple of lists): Tuple that contains lists of Input and Output objects.
                If one of them is not needed, leave list empty.
            finished_callback (Callable, optional): User-supplied function which will be called when the stream becomes inactive

        Returns:
            sd.Stream: Stream class to be used in a with statement.

        Raises:
            AttributeError: No audio device specified

        Example:
            ```python linenums="1"
            from asmu import ASetup

            setup = ASetup(setup_path)
            with setup.interface.stream((setup.inputs, setup.outputs)) as stream:
            
                # Start refill and processing threads

                input("Press ENTER to stop stream!")
                stream.stop()
            ```
        """
        try:
            self._init_sounddevice(io, sd.default)
        except AttributeError as exc:
            raise exc
        return sd.Stream(callback=self._callback, finished_callback=finished_callback)
    
    def stop_stream(self) -> None:
        """Stops the interface stream.

        Raises:
            sd.CallbackStop: Stop the sounddevice Stream.
        """
        raise sd.CallbackStop



"""Welcome to the API of the asmu module.
From the main [asmu][] module the classes for associated files can be imported.

Other classes and fuctions are grouped into the submodules:

- [asmu.io][]: For audio interface communication and definition of inputs and outputs.
- [asmu.generator][]: Signal chain objects creating or transmitting a signal.
- [asmu.analyzer][]: Signal chain objects consuming or recieving a signal.
"""

from asmu.afile import AFile
from asmu.asetup import ASetup

__all__ = ["AFile", "ASetup", "generator", "analyzer", "io"]



raise NotImplementedError
from PySide6 import QtGui
import pyqtgraph as pg
import numpy as np
import queue
import threading

pg.setConfigOption('background', 'k')
pg.setConfigOption('foreground', 'w')

class TimeView(pg.PlotWidget):
    def __init__(self, ioNames, ioColors, samplerate, queuesize:int=10, parent=None):
        super(TimeView, self).__init__(parent)
        # TODO give list of input classes instead?
        self.ioNames = ioNames
        self.ioQColors = [QtGui.QColor(ioColor) for ioColor in ioColors]

        self._samplerate = samplerate
        self.initPlots()

        self._q = queue.Queue(queuesize)

    def put_queue(self, data):
        self._q.put(data.copy())
    
    def process_queue(self):
        data = self._q.get(block=True, timeout=0.2)

        # plot buffer
        for idx, plot in enumerate(self._plots):
            plot.setData(data[:,idx])

    def start_process_thread(self, stream):
        def runner(self, stream):
            while stream.active:
                try:
                    self.process_queue()
                except queue.Empty:
                    pass
        threading.Thread(target=runner, args=(self, stream, )).start()


    def _get_fft(self) -> np.ndarray:
        return np.fft.rfft(self._q.get()*self._w, axis=0, norm="forward")*2/np.mean(self._w)

    def initPlots(self):
        # disable user interaction
        #self.setMouseEnabled(x=False, y=True)
        self.hideButtons()
        self.setMenuEnabled(False)

        # enable downsampling
        self.setDownsampling(auto=True, mode="subsample")

        # set ranges
        #self.setRange(yRange = (-10*self.pressureGrid, 10*self.pressureGrid), xRange = (-10*self.timeGrid*self.interfaceRate, 0), disableAutoRange=True)

        # set xTicks (with zero on the right)
        xAxis = self.getAxis("bottom")
        #xTicks = range(-10, 1, 1)

        #xAxis.setTicks([[(v*self.timeGrid*self.interfaceRate, f"{v*self.timeGrid:.2f}") for v in xTicks]])

        # set yTicks (with zero in the center)
        yAxis = self.getAxis("left")
        #yTicks = range(-20, 21, 1)
        #yAxis.setTicks([[(v*self.pressureGrid, f"{v*self.pressureGrid:.2f}") for v in yTicks ]])

        # set lables
        xAxis.setLabel("samples")
        yAxis.setLabel("amplitude")

        # add grid
        self.showGrid(x=True, y=True, alpha=0.2)

        # add legend
        self.addLegend()
        
        # create plots
        self._plots = []
        for name, color in zip(self.ioNames, self.ioQColors):
            plot = self.plot(pen=color, name=name)
            self._plots.append(plot)

class Corrview(pg.PlotWidget):
    def __init__(self, corrviewSetup, inputNames, inputQColors, outputName, interfaceRate, parent=None):
        super(Corrview, self).__init__(parent)
        self.corrviewSetup = corrviewSetup

        self.inputNames = inputNames
        self.inputQColors = inputQColors

        self.outputName = outputName

        self.interfaceRate = interfaceRate

        self.microphonePlots = []
        self.speakerPlots = []
        self.initPlots()

    @property
    def timeGrid(self):
        return float(self.corrviewSetup["timeGrid"])
    
    def updateView(self, corrs):
        for idx, plot in enumerate(self.inputPlots):
            plot.setData(corrs[idx])

    def initPlots(self):
        # disable user interaction
        #self.setMouseEnabled(x=False, y=True)
        #self.hideButtons()
        #self.setMenuEnabled(False)

        # enable downsampling
        self.setDownsampling(auto=True, mode="peak")

        xAxis = self.getAxis("bottom")
        yAxis = self.getAxis("left")

        # set lables
        xAxis.setLabel("delaytime (s)")
        yAxis.setLabel("match (%)")

        # add grid
        self.showGrid(x=True, y=True, alpha=0.2)

        # add legend
        self.addLegend()
        
        # create plots
        self.inputPlots = []
        for name, color in zip(self.inputNames, self.inputQColors):
            plot = self.plot(pen=color, name=f"{self.outputName}*{name}")
            self.inputPlots.append(plot)
        

    


import numpy as np
import scipy as sp
import queue
import threading

class Analyzer():
    def __init__(self, queuesize):
        """_summary_

        Args:
            queuesize (_type_): _description_
        """
        self._q = queue.Queue(queuesize)

    def reset(self):
        # process full queue
        while not self._q.empty():
            self._process_queue()

    def put_queue(self, data):
        self._q.put(data.copy())

    def _process_queue(self):
        raise NotImplementedError("Subclass has to define a process_queue function!")

    def start_process_thread(self, stream):
        def runner(self, stream):
            while stream.active:
                try:
                    self._process_queue()
                except queue.Empty:
                    pass
            # flush the remaining queue ater stream is closed
            while not self._q.empty():
                self._process_queue()
        threading.Thread(target=runner, args=(self, stream, )).start()

class Recorder(Analyzer):
    def __init__(self, signal, queuesize=10):
        """_summary_

        Args:
            signal (_type_): _description_
            queuesize (int, optional): _description_. Defaults to 10.
        """
        self._signal = signal
        super().__init__(queuesize)
        
    def _process_queue(self):
        self._signal.write(self._q.get(timeout=0.2))

class calcV(Analyzer):
    def __init__(self, samplerate, blocksize, queuesize: int=10):
        self._uks = []
        self._ks = []
        self.samplerate = samplerate
        self.blocksize = blocksize
        self._w = np.hanning(blocksize)
        super().__init__(queuesize)

    def _process_queue(self):
        ukm = self._get_fft()

        kmax = np.argmax(np.abs(ukm))
        self._uks.append(np.abs(ukm[kmax]))
        self._ks.append(kmax)

    def _get_fft(self) -> np.ndarray:
        # TODO: Move this to helper function
        return np.fft.rfft(self._q.get(timeout=0.2)*self._w[:, None], axis=0, norm="forward")*2/np.mean(self._w)

    def get_cV(self, Vp):
        self.reset()
        return Vp/np.mean(self._uks[2:]) # dont use the first three measurements

    def get_fV(self):
        freqs = np.fft.rfftfreq(self.blocksize, 1/self.samplerate)
        return np.interp(np.mean(self._ks[2:]), np.arange(freqs.size), freqs)
    
class calcPa(Analyzer):
    def __init__(self, samplerate, blocksize, queuesize: int=10):
        self._uks = []
        self._ks = []
        self.samplerate = samplerate
        self.blocksize = blocksize
        self._w = np.hanning(blocksize)
        super().__init__(queuesize)

    def _process_queue(self):
        ukm = self._get_fft()

        kmax = np.argmax(np.abs(ukm))
        self._uks.append(np.abs(ukm[kmax]))
        self._ks.append(kmax)

    def _get_fft(self) -> np.ndarray:
        # TODO: Move this to helper function
        return np.fft.rfft(self._q.get(timeout=0.2)*self._w[:, None], axis=0, norm="forward")*2/np.mean(self._w)

    def get_cPa(self, spl=94, Pa=None):
        self.reset()
        if Pa is not None:
            pcal = Pa
        else:
            p0 = 2e-5                           # [Pa]Rms
            p = p0 * 10 ** (spl / 20)           # [Pa]Rms
            pcal = p * np.sqrt(2)               # [Pa]
        

        return pcal/np.mean(self._uks[2:]) # dont use the first three measurements

    def get_fPa(self):
        freqs = np.fft.rfftfreq(self.blocksize, 1/self.samplerate)
        return np.interp(np.mean(self._ks[2:]), np.arange(freqs.size), freqs)
    
def get_corrs_sampleshifts(indata: np.ndarray, refdata: np.ndarray, burstLength: int) -> tuple[np.ndarray, np.ndarray]:
    corrs = []
    sampleshifts = []
    n = int(np.size(refdata))
    for cidx in range(np.ma.size(indata, axis=1)):
        v = indata[:, cidx]
        corr = sp.signal.correlate(v, refdata)
        corr /= np.max(corr)
        peaks, _ = sp.signal.find_peaks(corr, prominence=0.8, distance=burstLength, height=0.5)

        # shift correlation correctly
        corrs.append(corr[n-1:])
        sampleshifts.append(peaks[0]-(n-1))
    return np.array(corrs).T, np.array(sampleshifts)


            
        


import numpy as np
import scipy as sp
import threading
import queue

class Generator():
    def __init__(self, queuesize) -> None:
        """_summary_

        Args:
            queuesize (_type_): _description_
        """
        self._q = queue.Queue(queuesize)
        self._prefill_queue()

    def get_queue(self) -> queue.Queue:
        return self._q.get_nowait()

    def _prefill_queue(self) -> None:
        while self._q.full() is False:
           self._refill_queue(block=False)

    def _refill_queue(self, block=True) -> None:
        #if self._q.full() is False:
        data = self._gen()
        self._q.put(data, block=block, timeout=0.2)

    def start_refill_thread(self, stream):
        def runner(self, stream):
            while stream.active:
                try:
                    self._refill_queue()
                except queue.Full:
                    pass
        threading.Thread(target=runner, args=(self, stream, )).start()
    
    def _gen(self):
        raise NotImplementedError("Subclass has to define a _gen function!")

class Player(Generator):
    def __init__(self, asfile, blocksize, queuesize=10):
        """_summary_

        Args:
            asfile (_type_): _description_
            blocksize (_type_): _description_
            queuesize (int, optional): _description_. Defaults to 10.
        """
        self._asfile = asfile
        self._blocksize = blocksize
        super().__init__(queuesize)

    def _gen(self):
        data = self._asfile.read(self._blocksize, dtype="float32", always_2d=True)
        length = np.ma.size(data, axis=0)
        if data.size == 0:
            return None
        elif length < self._blocksize:
            pad = np.zeros((self._blocksize, self._asfile.channels))
            pad[:length, :] = data
            return pad
        else:
            return data
        
class Vector(Generator):
    def __init__(self, blocksize: int, vector:np.ndarray, queuesize=10):
        """_summary_

        Args:
            blocksize (int): _description_
            vector (np.ndarray): 2D Numpy vector to play with the shape [samples x channels].
            queuesize (int, optional): _description_. Defaults to 10.
        """
        self._blocksize = blocksize
        self._vector = np.copy(vector.astype(np.float32)).T # TODO - maybe improove this
        self._idx = 0
        super().__init__(queuesize)

    def _gen(self) -> np.ndarray:
        data = self._vector[self._idx:self._idx+self._blocksize, :]
        self._idx += self._blocksize

        length = np.ma.size(data, axis=0)
        if data.size == 0:
            return None
        elif length < self._blocksize:
            pad = np.zeros((self._blocksize, np.ma.size(data, axis=1)))
            pad[:length, :] = data
            return pad
        else:
            return data

class Chirp(Generator):
    def __init__(self, samplerate: int, blocksize: int, f0: float, f1: float, samples: int, queuesize: int=10) -> None:
        self._samplerate = samplerate
        self._blocksize = blocksize
        
        t = np.arange(0, samples) / samplerate
        self._chirp = sp.signal.chirp(t, f0=f0, t1=samples/samplerate, f1=f1, method="logarithmic")
        self._idx = 0
        super().__init__(queuesize)
    
    def _gen(self) -> np.ndarray:
        data = self._chirp[self._idx:self._idx+self._blocksize]
        self._idx += self._blocksize

        length = np.ma.size(data, axis=0)
        if data.size == 0:
            return None
        elif length < self._blocksize:
            pad = np.zeros(self._blocksize)
            pad[:length] = data
            return pad
        else:
            return data


class SineBurst(Generator):
    def __init__(self, samplerate: int, blocksize: int, f: float, n: int, loop: float=-1, queuesize: int=10) -> None:
        self._samplerate = samplerate
        self._blocksize = blocksize
        self._freq = f

        # generate wave parameters
        self._omegas = np.linspace(0, 2*np.pi*f*blocksize/samplerate, blocksize, endpoint=False, dtype=np.float32)
        self._phi = 0
        self._maxang = 2*np.pi*n
        self._loopang = 2*np.pi*loop*f

        super().__init__(queuesize)

    @property
    def burstlen(self):
        return self._samplerate/(2*np.pi*self._freq)*self._maxang

    def _gen(self) -> np.ndarray:
        ang = self._omegas + self._phi
        if self._loopang > 0: ang = np.mod(ang, self._loopang)
        sineburst = np.sin(ang)
        sineburst[ang > self._maxang] = 0
        self._phi += 2*np.pi*self._freq*self._blocksize/self._samplerate
        return sineburst
    
class SincBurst(Generator):
    def __init__(self, samplerate: int, blocksize: int, f: float, n: int, loop: float=-1, queuesize: int=10) -> None:
        self._samplerate = samplerate
        self._blocksize = blocksize
        self._freq = f

        # generate wave parameters
        self._omegas = np.linspace(0, 2*np.pi*f*blocksize/samplerate, blocksize, endpoint=False, dtype=np.float32)
        self._phi = 0
        self._maxang = 2*np.pi*n
        self._loopang = 2*np.pi*loop*f

        super().__init__(queuesize)

    @property
    def burstlen(self):
        return self._samplerate/(2*np.pi*self._freq)*self._maxang

    def _gen(self) -> np.ndarray:
        ang = self._omegas + self._phi
        if self._loopang > 0: ang = np.mod(ang, self._loopang)
        sincburst = np.sinc((ang-self._maxang/2)/np.pi)
        sincburst[ang > self._maxang] = 0
        self._phi += 2*np.pi*self._freq*self._blocksize/self._samplerate
        return sincburst
        
class WhiteNoise(Generator):
    def __init__(self, blocksize: int, queuesize: int=10) -> None:
        self._blocksize = blocksize

        super().__init__(queuesize)

    def _gen(self) -> np.ndarray:
        fwhite = np.fft.rfft(np.random.randn(self._blocksize))
        return np.clip(np.fft.irfft(fwhite)/4, -1, 1)
    
class PinkNoise(Generator):
    def __init__(self, blocksize: int, queuesize: int=10) -> None:
        self._blocksize = blocksize

        super().__init__(queuesize)

    def _gen(self) -> np.ndarray:
        fwhite = np.fft.rfft(np.random.randn(self._blocksize))

        # apply noise color and normalize for power
        f = np.fft.rfftfreq(self._blocksize)
        S = 1/np.where(f == 0, float('inf'), np.sqrt(f))
        S = S / np.sqrt(np.mean(S**2))
        fpink = fwhite * S

        return np.clip(np.fft.irfft(fpink)/4, -1, 1)

class Sine(Generator):
    def __init__(self, samplerate: int, blocksize: int, f: float, queuesize: int=10) -> None:
        self._samplerate = samplerate
        self._blocksize = blocksize
        self._freq = f

        # generate wave parameters
        self._omegas = np.linspace(0, 2*np.pi*f*blocksize/samplerate, blocksize, endpoint=False, dtype=np.float32)
        self._phi = 0

        super().__init__(queuesize)

    def _gen(self) -> np.ndarray:
        ang = self._omegas + self._phi
        self._phi += 2*np.pi*self._freq*self._blocksize/self._samplerate
        self._phi %= 2*np.pi
        return np.sin(ang)

                    
        


import json
import numpy as np
import os
from asmu.io import Input, Output, Interface
from typing import List

class ASetup():
    def __init__(self, asetup_path: str) -> None:
        """The ASetup class handles .asm_settings JSON files. 
        Loading and saving, initializing the respective io classes and properties.

        Args:
            asetup_path (str): Path to .asm_setup file.
        """
        self._path, file = os.path.split(asetup_path)
        self._name = file.replace(".asm_setup", "") # extract name without file ending
        self.load_file(asetup_path)

    @property
    def name(self) -> str:
        """ASetup file name without ending."""
        return str(self._name)

    @property
    def path(self) -> str:
        """ASetup file path."""
        return str(self._path)

    @property
    def inputs(self) -> List[Input]:
        """List of all current Input instances."""
        return self._inputs

    @property
    def outputs(self) -> List[Output]:
        """List of all current Output instances."""
        return self._outputs

    @property
    def interface(self) -> Interface:
        """Current instance of Interface."""
        return self._interface
    
    def _initio(self):
        self._interface = Interface(self._setup["interface"])

        self._inputs = []
        for in_setup in self._setup["inputs"]:
            self._inputs.append(Input(in_setup))

        self._outputs = []
        for out_setup in self._setup["outputs"]:
            self._outputs.append(Output(out_setup))

    def _load_npy(self):
        """Load numpy arrays from external files into setup dicts."""
        for io in self.inputs+self.outputs:
            npys = {}
            for (k, v) in io.io_setup.items():
                try:
                    if isinstance(v, str) and v.endswith(".npy"):
                        npys[k] = np.load(v)
                except FileNotFoundError:
                    pass
            for (k, v) in npys.items(): io.io_setup[k] = v

    def _save_npy(self):
        """Save numpy arrays to external files and remove them from setup dicts."""
        for io in self.inputs+self.outputs:
            paths = {}
            for (k, v) in io.io_setup.items():
                try:
                    if isinstance(v, np.ndarray):
                        path = f"{self.path}/{self.name}/{io.name}_{k}.npy"
                        np.save(path, v)
                        paths[k] = path
                except KeyError:
                    pass
            for (k, v) in paths.items(): io.io_setup[k] = v

    def load_file(self, asetup_path: str) -> None:
        """Load .asm_setup file and calibration files from associated folder.

        Args:
            asetup_path (str): Path to .asm_setup file.
        """
        with open(asetup_path, "r") as file:
            self._setup = json.load(file)

        self._initio()
        self._load_npy()

    def save_file(self, asetup_path: str=None) -> None:
        """Overwrite .asm_setup file or create copy at given asetup_path.

        Args:
            asetup_path (str, optional): Path to .asm_setup file.
        """
        # update path
        if asetup_path is not None:
            self._path, file = os.path.split(asetup_path)
            self._name = file.replace(".asm_setup", "") # extract name without file ending
        asetup_path = f"{self.path}/{self.name}.asm_setup"

        self._save_npy()

        # save .asm_setup file
        with open(asetup_path, 'w') as file:
            file.write(json.dumps(self._setup, sort_keys=True, indent=4, separators=(',', ': ')))

        self._load_npy()

        

    

    


import soundfile
import tempfile
import shutil
import os
import json
from datetime import datetime
import numpy as np


class AFile(soundfile.SoundFile):
    def __init__(self, afile_path:str, mode:str="r", samplerate:int=None, channels:int=None, temp:bool=False) -> None:
        """The AFile class handles .wav sound files. Metadata is used to store and load additional settings.
        It is recommended to open AFile's in a with statement.

        Args:
            afile_path (str): Path to .wav file.
            mode ({"r", "r+", "w", "w+"}, optional): File open mode.
            samplerate (int, optional): Samplerate, only used in write mode.
            channels (int, optional): Channel count, only used in write mode.
            temp (bool): Set file temporary (deleted on close). Only used in write mode.

        Example:
            ```python linenums="1" title="Load a file"
            from asmu import AFile

            with AFile(AFILE_PATH) as afile:
                print(afile.data.shape)
            ```
        """
        self._afile_path = afile_path
        self.settings = {}
        if mode == "r":
            super().__init__(afile_path, mode=mode)
            # load settings
            if self.comment:
                self.settings = json.loads(self.comment)
        else:
            _name = os.path.split(afile_path)[1].replace(".wav", "")
            if temp:
                self._tmp = tempfile.NamedTemporaryFile(prefix=_name, suffix='.wav', dir="", delete=True)
                super().__init__(self._tmp, mode=mode, samplerate=samplerate, channels=channels, subtype="PCM_24", format="WAV")
            else:
                super().__init__(afile_path, mode=mode, samplerate=samplerate, channels=channels, subtype="PCM_24", format="WAV")
            now = datetime.now()
            # set wav metadata
            self.title = _name
            self.date = now.strftime("%d/%m/%Y %H:%M:%S")
    
    # store settings as json string in metadata "comment"
    @property
    def settings(self) -> dict:
        """Additional JSON settings in the metadata's comment field."""
        return self._settings
    @settings.setter
    def settings(self, value: dict) -> None:
        self._settings = value

    @property
    def data(self) -> np.ndarray:
        """Data of AFile as numpy array of shape ("samples" x "channels")."""
        self.flush()
        self.seek(0)
        return self.read(dtype="float32", always_2d=True)
    
    def __exit__(self, *args):
        try:
            self.comment = json.dumps(self.settings)
        except soundfile.LibsndfileError:
            pass
        self.close()
